'''The LSSTMiddleManager is the only exported class: all the managers that
actually manage things report to it.  However, we expose some of the functions
from the group manager.
'''
from .middlemanager import LSSTMiddleManager
from .authmanager import check_membership, group_merge, deduplicate_groups

__all__ = [LSSTMiddleManager, check_membership,
           group_merge, deduplicate_groups]



'''This contains both the LSSTAuthManager class and a handful of
functions that are basically slightly-tailored versions of set
manipulation, designed for doing things with user group membership.
'''

import random
from tornado import gen
from ..utils import make_logger, list_duplicates, sanitize_dict


def check_membership(groups, allowed, forbidden, log=None):
    '''False if any of 'groups' are in 'forbidden'.  Otherwise true if
    either: 'allowed' is empty or at least one member of 'groups' is
    also in 'allowed'.  Pass a logger object in if you want it to emit
    debug messages.
    '''
    if forbidden:
        deny = list(set(forbidden) & set(groups))
        if deny:
            _maybe_log_debug(
                "User in forbidden group(s) '{}'".format(deny), log)
            return False
    if not allowed:
        _maybe_log_debug(
            "No list of allowed groups; allowing all groups.", log)
        return True
    intersection = list(set(allowed) & set(groups))
    if intersection:
        _maybe_log_debug(
            "User in allowed group(s) '{}'".format(intersection), log)
        return True
    return False


def _maybe_log_debug(message, log=None):
    if not log:
        return
    log.debug(message)


def group_merge(self, groups_1, groups_2):
    '''Merge two groups, renaming collisions.
    '''
    # First just merge them
    grps = []
    grps.extend(groups_1)
    grps.extend(groups_2)
    # Naive deduplication
    grps = list(set(grps))
    grpnames = [x.split(':', 1)[0] for x in grps]
    gnset = set(grpnames)
    # Check for need to do less-naive dedupe
    if len(gnset) != len(grpnames):
        # We have a collision
        grps = self._deduplicate_groups(grps, grpnames)
    return grps


def deduplicate_groups(grps):
    '''Rename groups with colliding names.
    '''
    grpsplits = [x.split(':', 1) for x in grps]
    grpnames = [x[0] for x in grpsplits]
    grpnums = [x[1] for x in grpsplits]
    flist = list_duplicates(grpnames)
    for name, positions in flist:
        i = 1
        for p in positions[1:]:
            # start with group_2 (leave the first one alone)
            i += 1
            grps[p] = grpnames[p] + "_" + str(i) + ":" + grpnums[p]
    return grps


class LSSTAuthManager(object):
    '''Class to hold LSST-specific authentication/authorization details
    and methods.

    Most of this was formerly held in the JupyterHub config as classes defined
    in '10-authenticator.py'.
    '''
    authenticator = None
    uid = None
    group_map = {}        # key is group name, value is group id
    auth_state = {}
    pod_env = {}

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        self.log.debug("Creating LSSTAuthManager.")
        self.parent = kwargs.pop('parent')
        self.authenticator = self.parent.authenticator

    def get_fake_gid(self):
        '''Use if we have strict_ldap_groups off, to assign GIDs to names
        with no matching Unix GID.  Since these will not appear as filesystem
        groups, being consistent with them isn't important.  We just need
        to make their GIDs something likely to not match anything real.

        There is a chance of collision, but it doesn't really matter.

        We do need to keep the no-GID groups around, though, because we might
        be using them to make options form or quota decisions (if we know we
        don't, we should turn on strict_ldap_groups).
        '''
        grpbase = 3E7
        grprange = 1E7
        igrp = random.randint(grpbase, (grpbase + grprange))
        return igrp

    def get_group_string(self):
        '''Convenience function for retrieving the group name-to-uid mapping
        list as a string suitable for passing to the spawned pod.
        '''
        return ','.join(["{}:{}".format(x, self.group_map[x])
                         for x in self.group_map])

    def get_pod_env(self):
        '''Return the authenticator-specific fields.
        '''
        return self.pod_env

    def parse_auth_state(self):
        '''Take the auth_state attribute and extract:
            * UID
            * Group/gid mappings
            * Possibly-authenticator-specific fields for pod environment

        and then store them in pod_env for the pod to pick up at spawn
         time. Our groups will be set if we authenticated, but if we got to
         the spawner page via a user that was already authenticated in the
         session DB (that is, you killed your pod and went away, but didn't
         log out, and then came back later while your session was still valid
         but the Hub had restarted), the authenticate() method in the spawner
         won't be run (since your user is still valid) but the fields won't
         be set (because the LSST managers are not persisted).  Hence the
         group mapping re-check, because to get around exactly this problem,
         each authenticator stores the group string in auth_state.
        '''
        self.log.debug("Parsing authentication state.")
        pod_env = {}
        ast = self.auth_state
        if not ast:
            raise RuntimeError(
                "Authenticator did not set manager's auth state!")
        cfg = self.parent.config
        authtype = cfg.authenticator_type
        if authtype == "github":
            gh_user = ast["github_user"]
            token = ast.get("access_token")
            gh_email = gh_user.get("email")
            gh_login = gh_user.get("login")
            gh_name = gh_user.get("name") or gh_login
            if gh_name:
                pod_env['GITHUB_NAME'] = gh_name
            if gh_login:
                pod_env['GITHUB_LOGIN'] = gh_login
            if gh_email:
                pod_env['GITHUB_EMAIL'] = gh_email
            pod_env['GITHUB_ACCESS_TOKEN'] = token
        elif authtype == "cilogon":
            email = ast.cilogon_user.get("email")  # ?
            if email:
                pod_env['GITHUB_EMAIL'] = email
        elif authtype == "jwt":
            claims = ast["claims"]
            token = ast["access_token"]
            email = claims.get("email")
            pod_env['ACCESS_TOKEN'] = token
            pod_env['GITHUB_EMAIL'] = email
        else:
            self.log.error("Authentication type {} unknown!".format(authtype))
        # These are generic
        self.uid = ast["uid"]
        if not self.uid:
            raise RuntimeError("Cannot determine user UID for pod spawn!")
        self.group_map = ast["group_map"]
        if not self.group_map:
            raise RuntimeError("Cannot determine user groups for pod spawn!")
        groupstr = self.get_group_string()
        pod_env['EXTERNAL_UID'] = str(self.uid)
        pod_env['EXTERNAL_GROUPS'] = groupstr
        self.pod_env = pod_env

    def dump(self):
        '''Return dict of contents for pretty-printing.
        '''
        pd = {"parent": str(self.parent),
              "uid": self.uid,
              "group_map": self.group_map,
              "auth_state": sanitize_dict(self.auth_state, ['access_token']),
              "pod_env": sanitize_dict(self.pod_env,
                                       ['ACCESS_TOKEN',
                                        'GITHUB_ACCESS_TOKEN'])}

        return pd



import argparse
import json
from ..config import LSSTConfig


def parse_args(cfg=LSSTConfig(),
               desc="Get list of Lab Images for display or prepulling",
               component="scanner"):
    '''Parse command-line arguments.
    '''
    parser = argparse.ArgumentParser(description=desc)
    dbg = cfg.debug
    lrh = cfg.lab_repo_host
    lro = cfg.lab_repo_owner
    lrn = cfg.lab_repo_name
    kex = cfg.prepuller_experimentals
    kdl = cfg.prepuller_dailies
    kwk = cfg.prepuller_weeklies
    krl = cfg.prepuller_releases
    ppc = cfg.prepuller_cachefile
    pto = cfg.prepuller_timeout
    ppn = cfg.prepuller_namespace
    lid = cfg.lab_uid
    cmd = cfg.prepuller_command
    component = component.lower()
    allowed_components = ["scanner", "reaper", "prepuller"]
    if component not in allowed_components:
        raise ValueError("Component {} not in {}!".format(component,
                                                          allowed_components))
    # These are larger for the reaper; there should be a number of not-
    #  displayed images we haven't destroyed yet.
    if component == "reaper":
        kex = cfg.reaper_keep_experimentals
        kdl = cfg.reaper_keep_dailies
        kwk = cfg.reaper_keep_weeklies

    parser.add_argument("-d", "--debug", action="store_true",
                        help="enable debugging [{}]".format(dbg),
                        default=dbg)
    parser.add_argument("-f", "--cachefile",
                        help="Cachefile for results [{}]".format(ppc),
                        default=ppc)
    parser.add_argument("-r", "--repo", "--repository",
                        help="Docker repository host [{}]".format(lrh),
                        default=lrh)
    parser.add_argument("-o", "--owner", "--organization", "--org",
                        help="repository owner [{}]".format(lro),
                        default=lro)
    parser.add_argument("-n", "--name",
                        help="repository name [{}]".format(lrn),
                        default=lrn)
    parser.add_argument("-q", "--dailies", "--daily", "--quotidian",
                        type=int,
                        help="# of daily builds to keep [{}]".format(kdl),
                        default=kdl)
    parser.add_argument("-w", "--weeklies", "--weekly", type=int,
                        help="# of weekly builds [{}]".format(kwk),
                        default=kwk)
    parser.add_argument("-e", "--experimentals", "--experimental",
                        "--exp", type=int,
                        help="# of experimental builds to keep [{}]".format(
                            kex),
                        default=kex)
    if component != "reaper":
        parser.add_argument("-b", "--releases", "--release", type=int,
                            help=("# of release builds to keep [{}]".format(
                                krl)),
                            default=krl)
    parser.add_argument("-p", "--port", help="Repository port [443 for" +
                        " secure, 80 for insecure]",
                        default=None)
    parser.add_argument("-i", "--insecure", "--no-tls", "--no-ssl",
                        help="Do not use TLS to connect [False]",
                        action='store_true',
                        default=False)
    if component != "reaper":
        parser.add_argument("-l", "--list", "--list-images", "--image-list",
                            help=("Use supplied comma-separated list in" +
                                  " addition to repo scan"),
                            default=None)
        parser.add_argument("-s", "--sort", "--sort-field", "--sort-by",
                            help="Field to sort results by [name]",
                            default="name")
        parser.add_argument("--recommended",
                            help=("Pull image with tag 'recommended' [True]"),
                            type=bool,
                            default=True)
    if component == "scanner":
        parser.add_argument("-j", "--json", help="Emit results as JSON, " +
                            " not sourceable shell fragment [False]",
                            action='store_true',
                            default=False)
    if component == "prepuller":
        parser.add_argument("--no-scan", action="store_true",
                            help=("Do not do repo scan (only useful in" +
                                  " conjunction with --list)"),
                            default=False)
        parser.add_argument("--command",
                            help=("JSON representation of command to run " +
                                  "when image is run as prepuller " +
                                  "[{}]".format(str(cmd))),
                            default=cmd)
        parser.add_argument("-t", "--timeout",
                            help=("Seconds allowed for process to complete " +
                                  "(-1 for no timeout) [{}]".format(pto)),
                            type=int,
                            default=pto)
        parser.add_argument("-u", "--uid",
                            help=("UID to run as [{}]".format(lid)),
                            type=int,
                            default=lid)
        parser.add_argument("--namespace",
                            help="Kubernetes namespace [{}]".format(ppn),
                            default=ppn)
    if component == "reaper":
        parser.add_argument("--dry-run", action='store_true',
                            help="Don't actually delete images",
                            default=False)
    results = parser.parse_args()
    results.path = ("/v2/repositories/" + results.owner + "/" +
                    results.name + "/tags/")
    if (hasattr(results, "list") and results.list):
        results.list = list(set(results.list.split(',')))
    if (hasattr(results, "command") and results.command and
            type(results.command) is str):
        results.command = json.loads(results.command)
    return results



#!/usr/bin/env python3
from .reaper import Reaper
from .parse_args import parse_args


def reaperstandalone():
    '''Standalone command for scanning repo.
    '''
    args = parse_args(desc="Remove obsolete lab images",
                      component="reaper")
    wilford_grimly = Reaper(host=args.repo, path=args.path,
                            owner=args.owner, name=args.name,
                            keep_dailies=args.dailies,
                            keep_weeklies=args.weeklies,
                            keep_experimentals=args.experimentals,
                            port=args.port, insecure=args.insecure,
                            cachefile=args.cachefile, dry_run=args.dry_run,
                            debug=args.debug)
    wilford_grimly.more_cowbell()


if __name__ == "__main__":
    reaperstandalone()



'''
JupyterHub Spawner to spawn user notebooks on a Kubernetes cluster in per-
user namespaces.

This module exports `NamespacedKubeSpawner` class, which is the spawner
implementation that should be used by JupyterHub.
'''

from jupyterhub.utils import exponential_backoff
from kubernetes import client
from kubernetes.client.rest import ApiException
from kubespawner import KubeSpawner
from kubespawner.clients import shared_client
from tornado import gen
from tornado.ioloop import IOLoop
from .multireflector import MultiNamespacePodReflector, EventReflector
from ..utils import make_logger


class MultiNamespacedKubeSpawner(KubeSpawner):
    '''Implement a JupyterHub spawner to spawn pods in a Kubernetes Cluster
    with per-user namespaces.
    '''

    rbac_api = None  # We need an RBAC client

    def __init__(self, *args, **kwargs):
        if not self.log:
            self.log = make_logger()
        super().__init__(*args, **kwargs)

        self.rbac_api = shared_client('RbacAuthorizationV1Api')

        selected_pod_reflector_classref = MultiNamespacePodReflector
        selected_event_reflector_classref = EventReflector
        self.namespace = self.get_user_namespace()

        main_loop = IOLoop.current()

        def on_pod_reflector_failure():
            self.log.critical("Pod reflector failed, halting Hub.")
            main_loop.stop()

        # Replace pod_reflector
        self.__class__.pod_reflector = selected_pod_reflector_classref(
            parent=self, namespace=self.namespace,
            on_failure=on_pod_reflector_failure
        )
        self.log.debug("Created new pod reflector: " +
                       "%r" % self.__class__.pod_reflector)
        # And event_reflector
        self.__class__.event_reflector = selected_event_reflector_classref(
            parent=self, namespace=self.namespace)

        # Defer event watcher startup to auth_state_hook to prevent
        #  initialization loop
        self.log.debug("MultiNamespaceSpawner initialized.")

    def auth_state_hook(self, auth_state):
        # Start the watchers.
        # Restart pod/event watcher
        self.log.debug("{} auth_state_hook firing.".format(__name__))
        self._start_watching_pods(replace=True)
        if self.events_enabled:
            self._start_watching_events(replace=True)

    @gen.coroutine
    def poll(self):
        '''
        Check if the pod is still running.

        Uses the same interface as subprocess.Popen.poll(): if the pod is
        still running, returns None.  If the pod has exited, return the
        exit code if we can determine it, or 1 if it has exited but we
        don't know how.  These are the return values JupyterHub expects.

        Note that a clean exit will have an exit code of zero, so it is
        necessary to check that the returned value is None, rather than
        just Falsy, to determine that the pod is still running.
        '''
        # have to wait for first load of data before we have a valid answer
        if not self.pod_reflector.first_load_future.done():
            yield self.pod_reflector.first_load_future
        data = self.pod_reflector.pods.get((self.namespace, self.pod_name),
                                           None)
        if data is not None:
            if data.status.phase == 'Pending':
                return None
            ctr_stat = data.status.container_statuses
            if ctr_stat is None:  # No status, no container (we hope)
                # This seems to happen when a pod is idle-culled.
                return 1
            for c in ctr_stat:
                # return exit code if notebook container has terminated
                if c.name == 'notebook':
                    if c.state.terminated:
                        # call self.stop to delete the pod
                        if self.delete_stopped_pods:
                            yield self.stop(now=True)
                        return c.state.terminated.exit_code
                    break
            # None means pod is running or starting up
            return None
        # pod doesn't exist or has been deleted
        return 1

    @gen.coroutine
    def _start(self):
        '''Start the user's pod.
        '''
        retry_times = 4  # Ad-hoc
        pod = yield self.get_pod_manifest()
        if self.modify_pod_hook:
            pod = yield gen.maybe_future(self.modify_pod_hook(self, pod))
        for i in range(retry_times):
            try:
                yield self.asynchronize(
                    self.api.create_namespaced_pod,
                    self.namespace,
                    pod,
                )
                break
            except ApiException as e:
                if e.status != 409:
                    # We only want to handle 409 conflict errors
                    self.log.exception("Failed for %s", pod.to_str())
                    raise
                self.log.info(
                    'Found existing pod %s, attempting to kill', self.pod_name)
                # TODO: this should show up in events
                yield self.stop(True)

                self.log.info(
                    'Killed pod %s, will try starting ' % self.pod_name +
                    'singleuser pod again')
        else:
            raise Exception(
                'Can not create user pod %s :' % self.pod_name +
                'already exists and could not be deleted')

        # we need a timeout here even though start itself has a timeout
        # in order for this coroutine to finish at some point.
        # using the same start_timeout here
        # essentially ensures that this timeout should never propagate up
        # because the handler will have stopped waiting after
        # start_timeout, starting from a slightly earlier point.
        try:
            yield exponential_backoff(
                lambda: self.is_pod_running(self.pod_reflector.pods.get(
                    (self.namespace, self.pod_name), None)),
                'pod/%s did not start in %s seconds!' % (
                    self.pod_name, self.start_timeout),
                timeout=self.start_timeout,
            )
        except TimeoutError:
            if self.pod_name not in self.pod_reflector.pods:
                # if pod never showed up at all,
                # restart the pod reflector which may have become disconnected.
                self.log.error(
                    "Pod %s never showed up in reflector;" % self.pod_name +
                    " restarting pod reflector."
                )
                self._start_watching_pods(replace=True)
            raise

        pod = self.pod_reflector.pods[(self.namespace, self.pod_name)]
        self.pod_id = pod.metadata.uid
        if self.event_reflector:
            self.log.debug(
                'pod %s events before launch: %s',
                self.pod_name,
                "\n".join(
                    [
                        "%s [%s] %s" % (event.last_timestamp,
                                        event.type, event.message)
                        for event in self.events
                    ]
                ),
            )
        return (pod.status.pod_ip, self.port)

    @gen.coroutine
    def stop(self, now=False):
        delete_options = client.V1DeleteOptions()

        if now:
            grace_seconds = 0
        else:
            # Give it some time, but not the default (which is 30s!)
            # FIXME: Move this into pod creation maybe?
            grace_seconds = 1

        delete_options.grace_period_seconds = grace_seconds
        self.log.info("Deleting pod %s", self.pod_name)
        try:
            yield self.asynchronize(
                self.api.delete_namespaced_pod,
                name=self.pod_name,
                namespace=self.namespace,
                body=delete_options,
                grace_period_seconds=grace_seconds,
            )
        except ApiException as e:
            if e.status == 404:
                self.log.warning(
                    "No pod %s to delete. Assuming already deleted.",
                    self.pod_name,
                )
            else:
                raise
        try:
            yield exponential_backoff(
                lambda: self.pod_reflector.pods.get((self.namespace,
                                                     self.pod_name), None) is
                None,
                'pod/%s did not disappear in %s seconds!' % (
                    self.pod_name, self.start_timeout),
                timeout=self.start_timeout,
            )
        except TimeoutError:
            self.log.error(
                "Pod %s did not disappear, " % self.pod_name +
                "restarting pod reflector")
            self._start_watching_pods(replace=True)
            raise



# Stolen from
# https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python


class Singleton(type):
    '''Singleton metaclass.  Create a Singleton class with:
        Class Foo(metaclass=Singleton)
    '''
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
